# Copyright (C) 2025 Embedl AB

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Union

import numpy as np
import onnx
import qai_hub as hub

from embedl_hub.core.context import RunType, experiment_context
from embedl_hub.core.quantization.quantization_config import QuantizationConfig
from embedl_hub.core.tuning.tuner import HubDataModule
from embedl_hub.tracking import log_param


class QuantizationError(RuntimeError):
    """Raised when Qualcomm AI Hub quantization job fails or times out."""


@dataclass
class QuantizationResult:
    model_path: Path
    job_id: str


def _generate_random_data(
    config: QuantizationConfig, input_name: str
) -> Dict[str, np.ndarray]:
    """
    Generate random calibration data for quantization.

    Args:
        config: QuantizationConfig containing img_size and num_samples.

    Returns:
        Dictionary with random calibration data.
    """
    config.num_samples = 1
    img_size = tuple(config.img_size)
    calibration_data = [np.random.rand(1, 3, *img_size).astype(np.float32)]

    return {input_name: calibration_data}


def _load_calibration_data(
    config: QuantizationConfig, input_name: str
) -> Dict[str, np.ndarray]:
    """
    Load calibration data from the specified path.

    Args:
        config: QuantizationConfig containing data_path and transforms.

    Returns:
        Dictionary with calibration data.
    """
    data_module = HubDataModule(config)
    data_module.prepare_data()
    data_module.setup("validate")

    calibration_data = []
    dataloader = (
        data_module.val_dataloader()
        if data_module.val_dataloader()
        else data_module.train_dataloader()
    )
    for batch_idx, (img, _) in enumerate(dataloader):
        if batch_idx >= config.num_samples:
            break
        calibration_data.append(img.numpy())

    return {input_name: calibration_data}


def collect_calibration_data(
    config: QuantizationConfig,
) -> Dict[str, np.ndarray]:
    """
    Collect calibration data for quantization.

    Args:
        config: QuantizationConfig containing data_path and transforms.

    Returns:
        Dictionary with calibration data.
    """
    onnx_model = onnx.load(config.model)
    input_names = [input.name for input in onnx_model.graph.input]
    assert len(input_names) == 1, "Model must have a single input."
    if config.data_path is None:
        return _generate_random_data(config, input_names[0])
    return _load_calibration_data(config, input_names[0])


def _save_quantized(quantized: "hub.Model", output_file: Path) -> Path:
    """
    Save the quantized artifact to a user-specified location or to the root
    directory with the name provided by Qualcomm AI Hub.

    Returns:
        Path to the saved model file.
    """
    output_file.parent.mkdir(parents=True, exist_ok=True)
    output_file = quantized.download(output_file)
    return output_file


def log_config_params(config: QuantizationConfig) -> None:
    """Log the configuration parameters for tracking."""

    def _format_img_size(value: Union[int, List[int]]) -> str:
        """Format image size for logging."""
        return str(value).strip("[]").replace(", ", "-").strip()

    log_param("num samples", str(config.num_samples))
    log_param("image size", _format_img_size(config.img_size))


def quantize_model(
    config: QuantizationConfig,
    project_name: str,
    experiment_name: str,
) -> QuantizationResult:
    """
    Submit an ONNX model to Qualcomm AI Hub and retrieve the compiled artifact.

    Args:
        config: QuantizationConfig containing model_path, data_path, and transforms.

    Returns:
        QuantizationResult with local path to compiled model.

    """

    with experiment_context(project_name, experiment_name, RunType.QUANTIZE):
        model_file = Path(config.model)
        if not model_file.exists():
            raise ValueError(f"Model not found: {model_file}")

        calibration_data = collect_calibration_data(config)
        log_config_params(config)

        try:
            job = hub.submit_quantize_job(
                model=model_file.as_posix(),
                weights_dtype=hub.QuantizeDtype.INT8,
                activations_dtype=hub.QuantizeDtype.INT8,
                calibration_data=calibration_data,
            )
        except Exception as error:
            raise QuantizationError(
                "Failed to submit quantization job."
            ) from error

        try:
            quantized = job.get_target_model()
        except Exception as error:
            raise QuantizationError(
                "Failed to download quantized model from Qualcomm AI Hub."
            ) from error
        local_path = _save_quantized(quantized, config.output_file)

        return QuantizationResult(
            model_path=local_path,
            job_id=job.job_id,
        )
